# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import typing

from datetime import datetime
from decimal import Decimal
from solana.publickey import PublicKey

from .accountinfo import AccountInfo
from .addressableaccount import AddressableAccount
from .context import Context
from .instrumentvalue import InstrumentValue
from .layouts import layouts
from .metadata import Metadata
from .observables import Disposable
from .tokens import Instrument, Token
from .version import Version
from .websocketsubscription import (
    WebSocketAccountSubscription,
    WebSocketSubscriptionManager,
)


# # 🥭 PriceCache class
#
# `PriceCache` stores a cached price.
#
class PriceCache:
    def __init__(self, price: Decimal, last_update: datetime) -> None:
        self.price: Decimal = price
        self.last_update: datetime = last_update

    @staticmethod
    def from_layout(layout: typing.Any) -> typing.Optional["PriceCache"]:
        if layout.last_update.timestamp() == 0:
            return None
        return PriceCache(layout.price, layout.last_update)

    def __str__(self) -> str:
        return f"« PriceCache [{self.last_update}] {self.price:,.20f} »"

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 RootBankCache class
#
# `RootBankCache` stores cached details of deposits and borrows.
#
class RootBankCache:
    def __init__(
        self, deposit_index: Decimal, borrow_index: Decimal, last_update: datetime
    ) -> None:
        self.deposit_index: Decimal = deposit_index
        self.borrow_index: Decimal = borrow_index
        self.last_update: datetime = last_update

    @staticmethod
    def from_layout(layout: typing.Any) -> typing.Optional["RootBankCache"]:
        if layout.last_update.timestamp() == 0:
            return None
        return RootBankCache(
            layout.deposit_index, layout.borrow_index, layout.last_update
        )

    def __str__(self) -> str:
        return f"« RootBankCache [{self.last_update}] {self.deposit_index:,.20f} / {self.borrow_index:,.20f} »"

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 PerpMarketCache class
#
# `PerpMarketCache` stores cached details of long and short funding.
#
class PerpMarketCache:
    def __init__(
        self, long_funding: Decimal, short_funding: Decimal, last_update: datetime
    ) -> None:
        self.long_funding: Decimal = long_funding
        self.short_funding: Decimal = short_funding
        self.last_update: datetime = last_update

    @staticmethod
    def from_layout(layout: typing.Any) -> typing.Optional["PerpMarketCache"]:
        if layout.last_update.timestamp() == 0:
            return None
        return PerpMarketCache(
            layout.long_funding, layout.short_funding, layout.last_update
        )

    def __str__(self) -> str:
        return f"« PerpMarketCache [{self.last_update}] {self.long_funding:,.20f} / {self.short_funding:,.20f} »"

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 MarketCache class
#
# `MarketCache` stores cached details of price, root bank, and perp market, for a particular market.
#
class MarketCache:
    def __init__(
        self,
        price: typing.Optional[PriceCache],
        root_bank: typing.Optional[RootBankCache],
        perp_market: typing.Optional[PerpMarketCache],
    ) -> None:
        self.price: typing.Optional[PriceCache] = price
        self.root_bank: typing.Optional[RootBankCache] = root_bank
        self.perp_market: typing.Optional[PerpMarketCache] = perp_market

    def adjusted_price(self, token: Instrument, quote_token: Token) -> InstrumentValue:
        if token == quote_token:
            # The price of 1 unit of the shared quote token is always 1.
            return InstrumentValue(quote_token, Decimal(1))

        if self.price is None:
            raise Exception(
                f"Could not find price index of basket token {token.symbol}."
            )

        price: Decimal = self.price.price
        decimals_difference = token.decimals - quote_token.decimals
        if decimals_difference != 0:
            adjustment = 10**decimals_difference
            price = price * adjustment

        return InstrumentValue(quote_token, price)

    def __str__(self) -> str:
        return f"""« MarketCache
    {self.price}
    {self.root_bank}
    {self.perp_market}
»"""

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 Cache class
#
# `Cache` stores cache details of prices, root banks and perp markets.
#
class Cache(AddressableAccount):
    def __init__(
        self,
        account_info: AccountInfo,
        version: Version,
        meta_data: Metadata,
        price_cache: typing.Sequence[typing.Optional[PriceCache]],
        root_bank_cache: typing.Sequence[typing.Optional[RootBankCache]],
        perp_market_cache: typing.Sequence[typing.Optional[PerpMarketCache]],
    ) -> None:
        super().__init__(account_info)
        self.version: Version = version

        self.meta_data: Metadata = meta_data
        self.price_cache: typing.Sequence[typing.Optional[PriceCache]] = price_cache
        self.root_bank_cache: typing.Sequence[
            typing.Optional[RootBankCache]
        ] = root_bank_cache
        self.perp_market_cache: typing.Sequence[
            typing.Optional[PerpMarketCache]
        ] = perp_market_cache

    @staticmethod
    def from_layout(
        layout: typing.Any, account_info: AccountInfo, version: Version
    ) -> "Cache":
        meta_data: Metadata = Metadata.from_layout(layout.meta_data)
        price_cache: typing.Sequence[typing.Optional[PriceCache]] = list(
            map(PriceCache.from_layout, layout.price_cache)
        )
        root_bank_cache: typing.Sequence[typing.Optional[RootBankCache]] = list(
            map(RootBankCache.from_layout, layout.root_bank_cache)
        )
        perp_market_cache: typing.Sequence[typing.Optional[PerpMarketCache]] = list(
            map(PerpMarketCache.from_layout, layout.perp_market_cache)
        )

        return Cache(
            account_info,
            version,
            meta_data,
            price_cache,
            root_bank_cache,
            perp_market_cache,
        )

    @staticmethod
    def parse(account_info: AccountInfo) -> "Cache":
        data = account_info.data
        if len(data) != layouts.CACHE.sizeof():
            raise Exception(
                f"Cache data length ({len(data)}) does not match expected size ({layouts.CACHE.sizeof()})"
            )

        layout = layouts.CACHE.parse(data)
        return Cache.from_layout(layout, account_info, Version.V1)

    @staticmethod
    def load(context: Context, address: PublicKey) -> "Cache":
        account_info = AccountInfo.load(context, address)
        if account_info is None:
            raise Exception(f"Cache account not found at address '{address}'")
        return Cache.parse(account_info)

    def subscribe(
        self,
        context: Context,
        websocketmanager: WebSocketSubscriptionManager,
        callback: typing.Callable[["Cache"], None],
    ) -> Disposable:
        subscription = WebSocketAccountSubscription(context, self.address, Cache.parse)
        websocketmanager.add(subscription)
        subscription.publisher.subscribe(on_next=callback)  # type: ignore[call-arg]

        return subscription

    def market_cache_for_index(self, index: int) -> MarketCache:
        return MarketCache(
            self.price_cache[index],
            self.root_bank_cache[index],
            self.perp_market_cache[index],
        )

    def __str__(self) -> str:
        def _render_list(
            items: typing.Sequence[typing.Any], stub: str
        ) -> typing.Sequence[str]:
            rendered = []
            for index, item in enumerate(items):
                rendered += [
                    f"{index}: {(item or stub)}".replace("\n", "\n            ")
                ]
            return rendered

        prices = "\n        ".join(_render_list(self.price_cache, "« No PriceCache »"))
        root_banks = "\n        ".join(
            _render_list(self.root_bank_cache, "« No RootBankCache »")
        )
        perp_markets = "\n        ".join(
            _render_list(self.perp_market_cache, "« No PerpMarketCache »")
        )
        return f"""« Cache [{self.version}] {self.address}
    {self.meta_data}
    Prices:
        {prices}
    Root Banks:
        {root_banks}
    Perp Markets:
        {perp_markets}
»"""



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://org-docs.friktion.fi/entropy/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/Friktion-Labs)
#   [Email](mailto:hello@blockworks.foundation)

import abc
import json
import logging
import os.path
import typing

from decimal import Decimal
from solana.publickey import PublicKey

from .constants import DATA_PATH, EntropyConstants
from .tokens import Instrument, Token


# # 🥭 InstrumentLookup class
#
# This class allows us to look up token symbols, names, decimals, and possibly mint addresses from Entropy
# and Solana static data.
#
# It's usually easiest to access it via the `Context` as `context.instrument_lookup`.
#
class InstrumentLookup(metaclass=abc.ABCMeta):
    def __init__(self) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)

    @abc.abstractmethod
    def find_by_symbol(self, symbol: str) -> typing.Optional[Instrument]:
        raise NotImplementedError(
            "InstrumentLookup.find_by_symbol() is not implemented on the base type."
        )

    @abc.abstractmethod
    def find_by_mint(self, mint: PublicKey) -> typing.Optional[Instrument]:
        raise NotImplementedError(
            "InstrumentLookup.find_by_mint() is not implemented on the base type."
        )

    def find_by_symbol_or_raise(self, symbol: str) -> Instrument:
        token = self.find_by_symbol(symbol)
        if token is None:
            raise Exception(f"Could not find token with symbol '{symbol}'.")

        return token

    def find_by_mint_or_raise(self, mint: PublicKey) -> Token:
        token = self.find_by_mint(mint)
        if token is None or not isinstance(token, Token):
            raise Exception(f"Could not find token with mint {mint}.")

        return token

    def __repr__(self) -> str:
        return f"{self}"

    def __str__(self) -> str:
        return """« InstrumentLookup »"""


# # 🥭 NullInstrumentLookup class
#
# This class is a simple stub `InstrumentLookup` that never returns a `Instrument`.
#
class NullInstrumentLookup(InstrumentLookup):
    def __init__(self) -> None:
        super().__init__()

    def find_by_symbol(self, symbol: str) -> typing.Optional[Instrument]:
        return None

    def find_by_mint(self, mint: PublicKey) -> typing.Optional[Instrument]:
        return None

    def __str__(self) -> str:
        return "« NullInstrumentLookup »"


# # 🥭 CompoundInstrumentLookup class
#
# This class allows multiple `InstrumentLookup` objects to be combined, returning the first valid lookup result
# found.
#
class CompoundInstrumentLookup(InstrumentLookup):
    def __init__(self, lookups: typing.Sequence[InstrumentLookup]) -> None:
        super().__init__()
        self.lookups: typing.Sequence[InstrumentLookup] = lookups

    def find_by_symbol(self, symbol: str) -> typing.Optional[Instrument]:
        for lookup in self.lookups:
            #     print(lookup)
            result = lookup.find_by_symbol(symbol)
            if result is not None:
                # print("found something", lookup)
                return result
        return None

    def find_by_mint(self, mint: PublicKey) -> typing.Optional[Instrument]:
        for lookup in self.lookups:
            result = lookup.find_by_mint(mint)
            if result is not None:
                return result
        return None

    def __str__(self) -> str:
        inner = "\n    ".join(
            [f"{item}".replace("\n", "\n    ") for item in self.lookups]
        )
        return f"""« CompoundInstrumentLookup
    {inner}
»"""


# # 🥭 NonSPLInstrumentLookup class
#
# This class allows us to look up non-SPL token data specifically from static data. This is useful for instruments
# that don't have an underlying SPL-token, such as ADA in ADA-PERP.
#
# You can load an `NonSPLInstrumentLookup` class by something like:
# ```
# with open("nonspl.instrumentlist.json") as json_file:
#     token_data = json.load(json_file)
#     token_lookup = NonSPLInstrumentLookup(token_data)
# ```
#
class NonSPLInstrumentLookup(InstrumentLookup):
    DefaultMainnetDataFilepath = os.path.join(DATA_PATH, "nonspl.instrumentlist.json")
    DefaultDevnetDataFilepath = os.path.join(
        DATA_PATH, "nonspl.instrumentlist.devnet.json"
    )

    def __init__(self, filename: str, token_data: typing.Dict[str, typing.Any]) -> None:
        super().__init__()
        self.filename: str = filename
        self.token_data: typing.Dict[str, typing.Any] = token_data

    def find_by_symbol(self, symbol: str) -> typing.Optional[Instrument]:
        # if symbol == "BTC^2":
        #     # Assumes BTC^2-PERP(6) will always have same amount of decimals as BTC(6)
        #     symbol = "BTC"
        for token in self.token_data["tokens"]:
            if Instrument.symbols_match(token["symbol"], symbol):
                return Instrument(
                    token["symbol"], token["name"], Decimal(token["decimals"])
                )
        # print("NonSPLInstrumentLookup found nothing")
        return None

    def find_by_mint(self, mint: PublicKey) -> typing.Optional[Instrument]:
        return None

    @staticmethod
    def load(filename: str) -> "NonSPLInstrumentLookup":
        with open(filename, encoding="utf-8") as json_file:
            token_data = json.load(json_file)
            return NonSPLInstrumentLookup(filename, token_data)

    def __str__(self) -> str:
        return f"« NonSPLInstrumentLookup [{self.filename}] »"


# # 🥭 IdsJsonTokenLookup class
#
# This class allows us to look up token data from our ids.json configuration file.
#
class IdsJsonTokenLookup(InstrumentLookup):
    def __init__(self, cluster_name: str, group_name: str) -> None:
        super().__init__()
        self.cluster_name: str = cluster_name
        self.group_name: str = group_name

    def find_by_symbol(self, symbol: str) -> typing.Optional[Token]:
        for group in EntropyConstants["groups"]:
            if (
                group["cluster"] == self.cluster_name
                and group["name"] == self.group_name
            ):
                for token in group["tokens"]:
                    if Instrument.symbols_match(token["symbol"], symbol):
                        return Token(
                            token["symbol"],
                            token["symbol"],
                            Decimal(token["decimals"]),
                            PublicKey(token["mintKey"]),
                        )
        # print("IdsJsonTokenLookup find_by_symbol found nothing")
        return None

    def find_by_mint(self, mint: PublicKey) -> typing.Optional[Token]:
        mint_str = str(mint)
        for group in EntropyConstants["groups"]:
            if (
                group["cluster"] == self.cluster_name
                and group["name"] == self.group_name
            ):
                for token in group["tokens"]:
                    if token["mintKey"] == mint_str:
                        return Token(
                            token["symbol"],
                            token["symbol"],
                            Decimal(token["decimals"]),
                            PublicKey(token["mintKey"]),
                        )
        return None

    def __str__(self) -> str:
        return f"« IdsJsonTokenLookup [{self.cluster_name}, {self.group_name}] »"


# # 🥭 SPLTokenLookup class
#
# This class allows us to look up token data specifically from Solana static data.
#
# The Solana static data is the [Solana token list](https://raw.githubusercontent.com/solana-labs/token-list/main/src/tokens/solana.tokenlist.json) provided by Serum.
#
# You can load an `SPLTokenLookup` class by something like:
# ```
# with open("solana.tokenlist.json") as json_file:
#     token_data = json.load(json_file)
#     token_lookup = SPLTokenLookup(token_data)
# ```
#
class SPLTokenLookup(InstrumentLookup):
    DefaultDataFilepath = os.path.join(DATA_PATH, "solana.tokenlist.json")
    DevnetDataFilepath = os.path.join(DATA_PATH, "solana.tokenlist.devnet.json")
    OverridesDataFilepath = os.path.join(DATA_PATH, "overrides.tokenlist.json")
    DevnetOverridesDataFilepath = os.path.join(
        DATA_PATH, "overrides.tokenlist.devnet.json"
    )

    def __init__(self, filename: str, token_data: typing.Dict[str, typing.Any]) -> None:
        super().__init__()
        self.filename: str = filename
        self.token_data: typing.Dict[str, typing.Any] = token_data

    def find_by_symbol(self, symbol: str) -> typing.Optional[Token]:
        for token in self.token_data["tokens"]:
            if Instrument.symbols_match(token["symbol"], symbol):
                return Token(
                    token["symbol"],
                    token["name"],
                    Decimal(token["decimals"]),
                    PublicKey(token["address"]),
                )
        # print("SPLTokenLookup found nothing")
        return None

    def find_by_mint(self, mint: PublicKey) -> typing.Optional[Token]:
        mint_string: str = str(mint)
        for token in self.token_data["tokens"]:
            if token["address"] == mint_string:
                return Token(
                    token["symbol"],
                    token["name"],
                    Decimal(token["decimals"]),
                    PublicKey(token["address"]),
                )

        return None

    @staticmethod
    def load(filename: str) -> "SPLTokenLookup":
        with open(filename, encoding="utf-8") as json_file:
            token_data = json.load(json_file)
            return SPLTokenLookup(filename, token_data)

    def __str__(self) -> str:
        return f"« SPLTokenLookup [{self.filename}] »"



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import logging
import entropy
import typing

from .element import Element
from ...modelstate import ModelState


# # 🥭 Chain class
#
# A `Chain` object takes a series of `Element`s and calls them in sequence to build a list of
# desired `Order`s.
#
# Only `Order`s returned from `process()` method are used as the final list of 'desired orders' for
# reconciling and possibly adding to the orderbook.
#
class Chain:
    def __init__(self, elements: typing.Sequence[Element]) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.elements: typing.Sequence[Element] = elements

    def process(
        self, context: entropy.Context, model_state: ModelState
    ) -> typing.Sequence[entropy.Order]:
        orders: typing.Sequence[entropy.Order] = []
        for element in self.elements:
            orders = element.process(context, model_state, orders)
        return orders

    def __repr__(self) -> str:
        return f"{self}"

    def __str__(self) -> str:
        elements = "\n    ".join(map(str, self.elements)) or "None"

        return f"""« Chain of {len(self.elements)} elements:
    {elements}
»"""



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://markets/) support is available at:
#   [Docs](https://docs.markets/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import rx
import rx.operators
import typing

from decimal import Decimal

from ...context import Context
from ...datetimes import utc_now
from ...loadedmarket import LoadedMarket
from ...observables import observable_pipeline_error_reporter
from ...oracle import (
    Oracle,
    OracleProvider,
    OracleSource,
    Price,
    SupportedOracleFeature,
)
from ...orders import OrderBook


# # 🥭 Market
#
# This file contains code for a 'market' oracle. A market can be one of:
# * Serum market
# * Entropy spot market
# * Entropy perp market
#


# # 🥭 MarketOracleConfidence constant
#
# Market doesn't provide a confidence value.
#
MarketOracleConfidence: Decimal = Decimal(0)


# # 🥭 MarketOracle class
#
# Implements the `Oracle` abstract base class specialised to the Entropy market.
#
class MarketOracle(Oracle):
    def __init__(self, market: LoadedMarket):
        name = f"Market Oracle for {market.fully_qualified_symbol}"
        super().__init__(name, market)
        self.loaded_market: LoadedMarket = market
        features: SupportedOracleFeature = SupportedOracleFeature.TOP_BID_AND_OFFER
        self.source: OracleSource = OracleSource("Market", name, features, market)

    def fetch_price(self, context: Context) -> Price:
        orderbook: OrderBook = self.loaded_market.fetch_orderbook(context)
        if orderbook.top_bid is None:
            raise Exception(
                f"[{self.source}] Cannot determine complete price data - no top bid"
            )
        top_bid = orderbook.top_bid.price

        if orderbook.top_ask is None:
            raise Exception(
                f"[{self.source}] Cannot determine complete price data - no top bid"
            )
        top_ask = orderbook.top_ask.price

        if orderbook.mid_price is None:
            raise Exception(
                f"[{self.source}] Cannot determine complete price data - no mid price"
            )
        mid_price = orderbook.mid_price

        return Price(
            self.source,
            utc_now(),
            self.market,
            top_bid,
            mid_price,
            top_ask,
            MarketOracleConfidence,
        )

    def to_streaming_observable(
        self, context: Context
    ) -> rx.core.typing.Observable[Price]:
        prices = rx.interval(1).pipe(
            rx.operators.observe_on(context.create_thread_pool_scheduler()),
            rx.operators.start_with(-1),
            rx.operators.map(lambda _: self.fetch_price(context)),
            rx.operators.catch(observable_pipeline_error_reporter),
            rx.operators.retry(),
        )
        return typing.cast(rx.core.typing.Observable[Price], prices)


# # 🥭 MarketOracleProvider class
#
# Implements the `OracleProvider` abstract base class specialised to the Entropy markets.
#
class MarketOracleProvider(OracleProvider):
    def __init__(self) -> None:
        super().__init__("Market Oracle Factory")

    def oracle_for_market(
        self, context: Context, market: LoadedMarket
    ) -> typing.Optional[Oracle]:
        return MarketOracle(market)

    def all_available_symbols(self, context: Context) -> typing.Sequence[str]:
        all_markets = context.market_lookup.all_markets()
        symbols: typing.List[str] = []
        for market in all_markets:
            symbols += [market.symbol]
        return symbols



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import argparse
import entropy
import typing

from decimal import Decimal

from .element import Element
from ...modelstate import ModelState


DEFAULT_SPREAD_RATIO = Decimal("0.01")
DEFAULT_POSITION_SIZE_RATIO = Decimal("0.01")


# # 🥭 RatiosElement class
#
# Ignores any input `Order`s (so probably best at the head of the chain). Builds orders using a spread
# ratio and a position size ratio.
#
class RatiosElement(Element):
    def __init__(
        self,
        order_type: entropy.OrderType,
        expire_seconds: typing.Optional[Decimal],
        match_limit: int,
        spread_ratios: typing.Sequence[Decimal],
        position_size_ratios: typing.Sequence[Decimal],
        from_bid_ask: bool,
    ) -> None:
        super().__init__()
        self.order_type: entropy.OrderType = order_type
        self.expire_seconds: typing.Optional[Decimal] = expire_seconds
        self.match_limit: int = match_limit
        self.spread_ratios: typing.Sequence[Decimal] = spread_ratios
        self.position_size_ratios: typing.Sequence[Decimal] = position_size_ratios
        self.from_bid_ask: bool = from_bid_ask

        if len(self.spread_ratios) == 0:
            raise Exception(
                "No spread ratios specified. Try the --ratios-spread parameter?"
            )

        if len(self.position_size_ratios) == 0:
            raise Exception(
                "No position-size ratios specified. Try the --ratios-position-size parameter?"
            )

        if len(self.spread_ratios) != len(self.position_size_ratios):
            raise Exception(
                "List of spread ratios and position size ratios must be the same length."
            )

    @staticmethod
    def add_command_line_parameters(parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--ratios-spread",
            type=Decimal,
            action="append",
            help="ratio to apply to the mid-price to create the BUY and SELL price (can be specified multiple times but every occurrance must have a matching --position-size-ratio occurrance)",
        )
        parser.add_argument(
            "--ratios-position-size",
            type=Decimal,
            action="append",
            help="ratio to apply to the available collateral to create the position size (can be specified multiple times but every occurrance must have a matching --spread-ratio occurrance)",
        )
        parser.add_argument(
            "--ratios-from-bid-ask",
            action="store_true",
            default=False,
            help="calculate ratios from bid or ask, not mid price (default: False, which will use the mid price)",
        )

    @staticmethod
    def from_command_line_parameters(args: argparse.Namespace) -> "RatiosElement":
        order_type: entropy.OrderType = args.order_type
        expire_seconds: typing.Optional[Decimal] = args.expire_seconds
        match_limit: int = args.match_limit or entropy.Order.DefaultMatchLimit
        spread_ratios: typing.Sequence[Decimal] = args.ratios_spread or [
            DEFAULT_SPREAD_RATIO
        ]
        position_size_ratios: typing.Sequence[Decimal] = args.ratios_position_size or [
            DEFAULT_POSITION_SIZE_RATIO
        ]
        from_bid_ask: bool = args.ratios_from_bid_ask
        return RatiosElement(
            order_type,
            expire_seconds,
            match_limit,
            spread_ratios,
            position_size_ratios,
            from_bid_ask,
        )

    def process(
        self,
        context: entropy.Context,
        model_state: ModelState,
        orders: typing.Sequence[entropy.Order],
    ) -> typing.Sequence[entropy.Order]:
        price: entropy.Price = model_state.price
        expiration = entropy.Order.build_absolute_expiration(self.expire_seconds)
        new_orders: typing.List[entropy.Order] = []
        for counter in range(len(self.spread_ratios)):
            position_size_ratio = self.position_size_ratios[counter]
            quote_value_to_risk = (
                model_state.inventory.available_collateral.value * position_size_ratio
            )
            base_position_size = quote_value_to_risk / price.mid_price

            spread_ratio = self.spread_ratios[counter]
            bid_price_base: Decimal = price.mid_price
            ask_price_base: Decimal = price.mid_price
            if self.from_bid_ask:
                bid_price_base = price.top_bid
                ask_price_base = price.top_ask

            bid: Decimal = bid_price_base - (bid_price_base * spread_ratio)
            ask: Decimal = ask_price_base + (ask_price_base * spread_ratio)

            bid_order = entropy.Order.from_values(
                entropy.Side.BUY,
                price=bid,
                quantity=base_position_size,
                order_type=self.order_type,
                expiration=expiration,
                match_limit=self.match_limit,
            )
            ask_order = entropy.Order.from_values(
                entropy.Side.SELL,
                price=ask,
                quantity=base_position_size,
                order_type=self.order_type,
                expiration=expiration,
                match_limit=self.match_limit,
            )
            self._logger.debug(
                f"""Desired orders:
    Bid: {bid_order}
    Ask: {ask_order}"""
            )
            new_orders += [bid_order, ask_order]

        return new_orders

    def __str__(self) -> str:
        spread_ratios = ", ".join(map(str, self.spread_ratios)) or "None"
        position_size_ratios = ", ".join(map(str, self.position_size_ratios)) or "None"
        from_description = "from bid/ask" if self.from_bid_ask else "from mid price"
        return f"« RatiosElement using ratios - spread(s): {spread_ratios} {from_description}, position size(s): {position_size_ratios} »"



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import logging
import numpy
import pandas
import os
import os.path
import requests
import time
import typing

from datetime import datetime, timedelta
from dateutil import parser
from decimal import Decimal
from solana.publickey import PublicKey

from .account import Account
from .context import Context


# # 🥭 TradeHistory class
#
# Downloads and unifies trade history data.
#
class TradeHistory:
    COLUMNS = [
        "Timestamp",
        "Market",
        "Side",
        "MakerOrTaker",
        "Change",
        "Price",
        "Quantity",
        "Fee",
        "SequenceNumber",
        "FeeTier",
        "MarketType",
        "Counterparty",
        "OrderId",
        "ClientId",
    ]

    __perp_column_name_mapper = {
        "loadTimestamp": "Timestamp",
        "seqNum": "SequenceNumber",
        "price": "Price",
        "quantity": "Quantity",
    }

    __spot_column_name_mapper = {
        "loadTimestamp": "Timestamp",
        "seqNum": "SequenceNumber",
        "price": "Price",
        "size": "Quantity",
        "side": "Side",
        "feeCost": "Fee",
        "feeTier": "FeeTier",
        "orderId": "OrderId",
        "clientOrderId": "ClientId",
    }

    __decimal_spot_columns = [
        "openOrderSlot",
        "feeTier",
        "nativeQuantityReleased",
        "nativeQuantityPaid",
        "nativeFeeOrRebate",
        "orderId",
        "clientOrderId",
        "source",
        "seqNum",
        "baseTokenDecimals",
        "quoteTokenDecimals",
        "price",
        "feeCost",
        "size",
    ]

    __decimal_perp_columns = [
        "seqNum",
        "makerFee",
        "takerFee",
        "makerOrderId",
        "takerOrderId",
        "price",
        "quantity",
        "makerClientOrderId",
        "takerClientOrderId",
    ]

    __column_converters = {
        "Timestamp": lambda value: parser.parse(value),
        "SequenceNumber": lambda value: Decimal(value),
        "Price": lambda value: Decimal(value),
        "Change": lambda value: Decimal(value),
        "Quantity": lambda value: Decimal(value),
        "Fee": lambda value: Decimal(value),
        "FeeTier": lambda value: Decimal(value),
        "OrderId": lambda value: Decimal(value),
    }

    def __init__(self, seconds_pause_between_rest_calls: int = 1) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.__seconds_pause_between_rest_calls: int = seconds_pause_between_rest_calls
        self.__trades: pandas.DataFrame = pandas.DataFrame(columns=TradeHistory.COLUMNS)

    @staticmethod
    def __market_lookup(context: Context) -> typing.Callable[[pandas.Series], str]:
        def __safe_lookup(row: pandas.Series) -> str:
            address: PublicKey = PublicKey(row["address"])
            market = context.market_lookup.find_by_address(address)
            if market is None:
                raise Exception(f"No market found with address {address}")
            return market.symbol

        return __safe_lookup

    @staticmethod
    def __download_json(url: str) -> typing.Any:
        response = requests.get(url)
        response.raise_for_status()
        return response.json()

    @staticmethod
    def __download_all_perps(context: Context, account: Account) -> pandas.DataFrame:
        url = f"https://event-history-api.herokuapp.com/perp_trades/{account.address}?page=all"
        data = TradeHistory.__download_json(url)
        trades: pandas.DataFrame = TradeHistory.__perp_data_to_dataframe(
            context, account, data
        )

        return trades

    @staticmethod
    def __download_updated_perps(
        context: Context,
        account: Account,
        newer_than: typing.Optional[datetime],
        seconds_pause_between_rest_calls: int,
    ) -> pandas.DataFrame:
        trades: pandas.DataFrame = pandas.DataFrame(columns=TradeHistory.COLUMNS)
        page: int = 0
        complete: bool = False
        while not complete:
            page += 1
            url = f"https://event-history-api.herokuapp.com/perp_trades/{account.address}?page={page}"
            data = TradeHistory.__download_json(url)
            frame: pandas.DataFrame = TradeHistory.__perp_data_to_dataframe(
                context, account, data
            )
            if len(frame) == 0:
                complete = True
            else:
                trades = pandas.concat([trades, frame])
                if (newer_than is not None) and (
                    frame.loc[frame.index[-1], "Timestamp"] < newer_than
                ):
                    complete = True
                else:
                    time.sleep(seconds_pause_between_rest_calls)

        return trades

    @staticmethod
    def __perp_data_to_dataframe(
        context: Context, account: Account, data: typing.Any
    ) -> pandas.DataFrame:
        # Perp data is an array of JSON packages like:
        # {
        #     "loadTimestamp": "2021-09-02T10:54:56.000Z",
        #     "address": <PUBLIC-KEY-STRING>,
        #     "seqNum": "2831",
        #     "makerFee": "0",
        #     "takerFee": "0.0004999999999988347",
        #     "takerSide": "sell",
        #     "maker": <PUBLIC-KEY-STRING>,
        #     "makerOrderId": <BIG-INT>,
        #     "taker": <PUBLIC-KEY-STRING>,
        #     "takerOrderId": <BIG-INT>,
        #     "price": "50131.9",
        #     "quantity": "0.019",
        #     "makerClientOrderId": "1646845270119",
        #     "takerClientOrderId": "1646845251729"
        # },
        def __side_lookup(row: pandas.Series) -> str:
            if row["MakerOrTaker"] == "taker":
                return str(row["takerSide"])
            elif row["takerSide"] == "buy":
                return "sell"
            else:
                return "buy"

        def __fee_calculator(row: pandas.Series) -> Decimal:
            price: Decimal = row["Price"]
            quantity: Decimal = row["Quantity"]
            fee_rate: Decimal
            if row["MakerOrTaker"] == "maker":
                fee_rate = row["makerFee"]
            else:
                fee_rate = row["takerFee"]
            return price * quantity * -fee_rate

        if len(data["data"]) <= 1:
            return pandas.DataFrame(columns=TradeHistory.COLUMNS)

        trade_data = data["data"][:-1]
        for trade in trade_data:
            for column_name in TradeHistory.__decimal_perp_columns:
                trade[column_name] = Decimal(trade[column_name])

        frame = pandas.DataFrame(trade_data).rename(
            mapper=TradeHistory.__perp_column_name_mapper, axis=1, copy=True
        )
        frame["Timestamp"] = frame["Timestamp"].apply(
            lambda timestamp: parser.parse(timestamp).replace(microsecond=0)
        )
        frame["Market"] = frame.apply(TradeHistory.__market_lookup(context), axis=1)
        frame["MarketType"] = "perp"

        this_address = f"{account.address}"
        frame["MakerOrTaker"] = frame["maker"].apply(
            lambda addy: "maker" if addy == this_address else "taker"
        )

        frame["FeeTier"] = -1
        frame["Fee"] = frame.apply(__fee_calculator, axis=1)
        frame["Side"] = frame.apply(__side_lookup, axis=1)
        frame["Value"] = frame["Price"] * frame["Quantity"]
        frame["Change"] = (
            frame["Value"].where(frame["Side"] == "sell", other=-frame["Value"])
            + frame["Fee"]
        )
        frame["OrderId"] = numpy.where(
            frame["MakerOrTaker"] == "maker",
            frame["makerOrderId"],
            frame["takerOrderId"],
        )
        frame["ClientId"] = numpy.where(
            frame["MakerOrTaker"] == "maker",
            frame["makerClientOrderId"],
            frame["takerClientOrderId"],
        )
        frame["Counterparty"] = numpy.where(
            frame["MakerOrTaker"] == "maker",
            frame["taker"],
            frame["maker"],
        )

        return frame[TradeHistory.COLUMNS]

    @staticmethod
    def __download_all_spots(context: Context, account: Account) -> pandas.DataFrame:
        trades: pandas.DataFrame = pandas.DataFrame(columns=TradeHistory.COLUMNS)
        for spot_open_orders_address in account.spot_open_orders:
            url = f"https://event-history-api.herokuapp.com/trades/open_orders/{spot_open_orders_address}?page=all"
            data = TradeHistory.__download_json(url)
            frame = TradeHistory.__spot_data_to_dataframe(context, account, data)
            trades = pandas.concat([trades, frame])

        return trades

    @staticmethod
    def __download_updated_spots(
        context: Context,
        account: Account,
        newer_than: typing.Optional[datetime],
        seconds_pause_between_rest_calls: int,
    ) -> pandas.DataFrame:
        trades: pandas.DataFrame = pandas.DataFrame(columns=TradeHistory.COLUMNS)
        for spot_open_orders_address in account.spot_open_orders:
            page: int = 0
            complete: bool = False
            while not complete:
                page += 1
                url = f"https://event-history-api.herokuapp.com/trades/open_orders/{spot_open_orders_address}?page={page}"
                data = TradeHistory.__download_json(url)
                frame = TradeHistory.__spot_data_to_dataframe(context, account, data)

                if len(frame) == 0:
                    complete = True
                else:
                    trades = pandas.concat([trades, frame])
                    earliest_in_frame = frame.loc[frame.index[-1], "Timestamp"]
                    if (newer_than is not None) and (earliest_in_frame < newer_than):
                        complete = True
                    else:
                        time.sleep(seconds_pause_between_rest_calls)

        return trades

    @staticmethod
    def __spot_data_to_dataframe(
        context: Context, account: Account, data: typing.Any
    ) -> pandas.DataFrame:
        # Spot data is an array of JSON packages like:
        # {
        #     "loadTimestamp": "2021-10-05T16:04:50.717Z",
        #     "address": <PUBLIC-KEY-STRING>,
        #     "programId": <PUBLIC-KEY-STRING>,
        #     "baseCurrency": "SOL",
        #     "quoteCurrency": "USDC",
        #     "fill": true,
        #     "out": false,
        #     "bid": true,
        #     "maker": true,
        #     "openOrderSlot": "0",
        #     "feeTier": "4",
        #     "nativeQuantityReleased": "3000000000",
        #     "nativeQuantityPaid": "487482712",
        #     "nativeFeeOrRebate": "146288",
        #     "orderId": <BIG-INT>,
        #     "openOrders": <PUBLIC-KEY-STRING>,
        #     "clientOrderId": <BIG-INT>,
        #     "uuid": <LONG-OPAQUE-UUID-STRING>,
        #     "source": "2",
        #     "seqNum": "24827175",
        #     "baseTokenDecimals": 9,
        #     "quoteTokenDecimals": 6,
        #     "side": "buy",
        #     "price": 162.543,
        #     "feeCost": -0.146288,
        #     "size": 3
        # }
        if len(data["data"]) == 0:
            return pandas.DataFrame(columns=TradeHistory.COLUMNS)
        else:
            trade_data = data["data"]
            for trade in trade_data:
                for column_name in TradeHistory.__decimal_spot_columns:
                    trade[column_name] = Decimal(trade[column_name])

            frame = pandas.DataFrame(trade_data).rename(
                mapper=TradeHistory.__spot_column_name_mapper, axis=1, copy=True
            )
            frame["Timestamp"] = frame["Timestamp"].apply(
                lambda timestamp: parser.parse(timestamp).replace(microsecond=0)
            )
            frame["Market"] = frame.apply(TradeHistory.__market_lookup(context), axis=1)
            frame["MakerOrTaker"] = numpy.where(frame["maker"], "maker", "taker")
            frame["Fee"] = -frame["Fee"]
            frame["Value"] = frame["Price"] * frame["Quantity"]
            frame["Change"] = (
                frame["Value"].where(frame["Side"] == "sell", other=-frame["Value"])
                + frame["Fee"]
            )
            frame["MarketType"] = "spot"
            frame["Counterparty"] = None

            return frame[TradeHistory.COLUMNS]

    @property
    def trades(self) -> pandas.DataFrame:
        return self.__trades.copy(deep=True)

    def download_latest(
        self, context: Context, account: Account, cutoff: datetime
    ) -> None:
        # Go back further than we need to so we can be sure we're not skipping any trades due to race conditions.
        # We remove duplicates a few lines further down.
        self._logger.info(f"Downloading spot trades from {cutoff}")
        spot: pandas.DataFrame = TradeHistory.__download_updated_spots(
            context, account, cutoff, self.__seconds_pause_between_rest_calls
        )
        self._logger.info(f"Downloading perp trades from {cutoff}")
        perp: pandas.DataFrame = TradeHistory.__download_updated_perps(
            context, account, cutoff, self.__seconds_pause_between_rest_calls
        )

        all_trades: pandas.DataFrame = pandas.concat([self.__trades, spot, perp])
        all_trades = all_trades[all_trades["Timestamp"] >= cutoff]

        distinct_trades = all_trades.drop_duplicates()
        sorted_trades = distinct_trades.sort_values(
            ["Timestamp", "Market", "SequenceNumber"], axis=0, ascending=True
        )
        self._logger.info(
            f"Download complete. Data contains {len(sorted_trades)} trades."
        )
        self.__trades = sorted_trades

    def update(self, context: Context, account: Account) -> None:
        latest_trade: typing.Optional[datetime] = (
            self.__trades.loc[self.__trades.index[-1], "Timestamp"]
            if len(self.__trades) > 0
            else None
        )
        spot: pandas.DataFrame
        perp: pandas.DataFrame
        if latest_trade is None:
            self._logger.info("Downloading all spot trades.")
            spot = TradeHistory.__download_all_spots(context, account)
            self._logger.info("Downloading all perp trades.")
            perp = TradeHistory.__download_all_perps(context, account)
        else:
            # Go back further than we need to so we can be sure we're not skipping any trades due to race conditions.
            # We remove duplicates a few lines further down.
            cutoff_safety_margin: timedelta = timedelta(hours=1)
            cutoff: datetime = latest_trade - cutoff_safety_margin
            self._logger.info(
                f"Downloading spot trades from {cutoff}, {cutoff_safety_margin} before latest stored trade at {latest_trade}"
            )
            spot = TradeHistory.__download_updated_spots(
                context, account, cutoff, self.__seconds_pause_between_rest_calls
            )
            self._logger.info(
                f"Downloading perp trades from {cutoff}, {cutoff_safety_margin} before latest stored trade at {latest_trade}"
            )
            perp = TradeHistory.__download_updated_perps(
                context, account, cutoff, self.__seconds_pause_between_rest_calls
            )

        all_trades = pandas.concat([self.__trades, spot, perp])
        distinct_trades = all_trades.drop_duplicates()
        sorted_trades = distinct_trades.sort_values(
            ["Timestamp", "Market", "SequenceNumber"], axis=0, ascending=True
        )
        self._logger.info(
            f"Download complete. Data contains {len(sorted_trades)} trades."
        )
        self.__trades = sorted_trades

    def load(self, filename: str, ok_if_missing: bool = False) -> None:
        if not os.path.isfile(filename):
            if not ok_if_missing:
                raise Exception(f"File {filename} does not exist or is not a file.")
        else:
            existing = pandas.read_csv(
                filename,
                float_precision="round_trip",
                converters=TradeHistory.__column_converters,
            )

            self.__trades = pandas.concat([self.__trades, existing])

    def save(self, filename: str) -> None:
        self.__trades.to_csv(filename, index=False, mode="w")

    def __str__(self) -> str:
        return f"« TradeHistory containing {len(self.__trades)} trades »"

    def __repr__(self) -> str:
        return f"{self}"



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import enum
import logging
import typing

from decimal import Decimal, ROUND_DOWN, ROUND_UP
from solana.publickey import PublicKey

from .constants import SOL_DECIMALS, SOL_MINT_ADDRESS


class RoundDirection(enum.Enum):
    DOWN = enum.auto()
    UP = enum.auto()

    def to_decimal_direction(self) -> str:
        if self == RoundDirection.UP:
            return ROUND_UP
        else:
            return ROUND_DOWN

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"{self}"


class Instrument:
    def __init__(self, symbol: str, name: str, decimals: Decimal) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.symbol: str = symbol.upper()
        self.name: str = name
        self.decimals: Decimal = decimals

    @staticmethod
    def symbols_match(symbol1: str, symbol2: str) -> bool:
        return symbol1.upper() == symbol2.upper()

    def round(
        self, value: Decimal, direction: RoundDirection = RoundDirection.DOWN
    ) -> Decimal:
        # places should be a number like 0.000001
        places = Decimal(1) / (Decimal(10) ** self.decimals)
        return value.quantize(places, rounding=direction.to_decimal_direction())

    def shift_to_decimals(self, value: Decimal) -> Decimal:
        divisor = Decimal(10**self.decimals)
        shifted = value / divisor
        return shifted

    def shift_to_native(self, value: Decimal) -> Decimal:
        multiplier = Decimal(10**self.decimals)
        shifted = value * multiplier
        return round(shifted, 0)

    def symbol_matches(self, symbol: str) -> bool:
        return self.symbol.upper() == symbol.upper()

    # Instruments are equal if they have the same normalised symbol.
    def __eq__(self, other: typing.Any) -> bool:
        if isinstance(other, Instrument):
            return self.symbol == other.symbol
        return False

    def __str__(self) -> str:
        return f"« Instrument [{self.symbol}] '{self.name}' »"

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 Token class
#
# `Token` defines aspects common to every token.
#
class Token(Instrument):
    def __init__(
        self, symbol: str, name: str, decimals: Decimal, mint: PublicKey
    ) -> None:
        super().__init__(symbol, name, decimals)
        self.mint: PublicKey = mint

    def ensure(uncertain_token: Instrument) -> "Token":
        if isinstance(uncertain_token, Token):
            return uncertain_token
        raise Exception(
            f"Instrument {uncertain_token} cannot be converted to SPL Token"
        )

    @staticmethod
    def find_by_symbol(values: typing.Sequence["Token"], symbol: str) -> "Token":
        found = [value for value in values if value.symbol_matches(symbol)]
        if len(found) == 0:
            raise Exception(f"Token '{symbol}' not found in token values: {values}")

        if len(found) > 1:
            raise Exception(
                f"Token '{symbol}' matched multiple tokens in values: {values}"
            )

        return found[0]

    @staticmethod
    def find_by_mint(values: typing.Sequence["Token"], mint: PublicKey) -> "Token":
        found = [value for value in values if value.mint == mint]
        if len(found) == 0:
            raise Exception(f"Token '{mint}' not found in token values: {values}")

        if len(found) > 1:
            raise Exception(
                f"Token '{mint}' matched multiple tokens in values: {values}"
            )

        return found[0]

    # Tokens are equal if they have the same mint address.
    def __eq__(self, other: typing.Any) -> bool:
        if isinstance(other, Token):
            return self.mint == other.mint
        return False

    def __str__(self) -> str:
        return f"« Token [{self.symbol}] '{self.name}' [{self.mint} ({self.decimals} decimals)] »"


# # 🥭 SolToken object
#
# It's sometimes handy to have a `Token` for SOL, but SOL isn't actually a token and can't appear in baskets. This object defines a special case for SOL.
#
SolToken = Token("SOL", "Pure SOL", SOL_DECIMALS, SOL_MINT_ADDRESS)



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import abc
import logging
import typing

from decimal import Decimal

from .account import Account
from .context import Context
from .group import Group
from .marketoperations import MarketOperations
from .porcelain import operations
from .tokens import Instrument, Token
from .instrumentvalue import InstrumentValue
from .wallet import Wallet


# # 🥭 WalletBalancer
#
# This notebook deals with balancing a wallet after processing liquidations, so that it has
# appropriate funds for the next liquidation.
#
# We want to be able to maintain liquidity in our wallet. For instance if there are a lot of
# ETH shorts being liquidated, we'll need to supply ETH, but what do we do when we run out
# of ETH and there are still liquidations to perform?
#
# We 'balance' our wallet tokens, buying or selling or swapping them as required.
#

# # 🥭 Target Balances
#
# To be able to maintain the right balance of tokens, we need to know what the right
# balance is. Different people have different opinions, and we don't all have the same
# value in our liquidator accounts, so we need a way to allow whoever is running the
# liquidator to specify what the target token balances should be.
#
# There are two possible approaches to specifying the target value:
# * A 'fixed' value, like 10 ETH
# * A 'percentage' value, like 20% ETH
#
# Percentage is trickier, because to come up with the actual target we need to take into
# account the wallet value and the current price of the target token.
#
# The way this all hangs together is:
# * A parser parses string values (probably from a command-line) into `TargetBalance`
#   objects.
# * There are two types of `TargetBalance` objects - `FixedTargetBalance` and
#   `PercentageTargetBalance`.
# * To get the actual `InstrumentValue` for balancing, the `TargetBalance` must be 'resolved'
#   by calling `resolve()` with the appropriate token price and wallet value.
#

# # 🥭 TargetBalance class
#
# This is the abstract base class for our target balances, to allow them to be treated polymorphically.
#
class TargetBalance(metaclass=abc.ABCMeta):
    def __init__(self, symbol: str) -> None:
        self.symbol = symbol

    @abc.abstractmethod
    def resolve(
        self, instrument: Instrument, current_price: Decimal, total_value: Decimal
    ) -> InstrumentValue:
        raise NotImplementedError(
            "TargetBalance.resolve() is not implemented on the base type."
        )

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 FixedTargetBalance class
#
# This is the simple case, where the `FixedTargetBalance` object contains enough information on its own to build the resolved `InstrumentValue` object.
#
class FixedTargetBalance(TargetBalance):
    def __init__(self, symbol: str, value: Decimal) -> None:
        super().__init__(symbol)
        self.value = value

    def resolve(
        self, instrument: Instrument, current_price: Decimal, total_value: Decimal
    ) -> InstrumentValue:
        return InstrumentValue(instrument, self.value)

    def __str__(self) -> str:
        return f"""« FixedTargetBalance [{self.value} {self.symbol}] »"""


# # 🥭 PercentageTargetBalance
#
# This is the more complex case, where the target is a percentage of the total wallet
# balance.
#
# So, to actually calculate the right target, we need to know the total wallet balance and
# the current price. Once we have those the calculation is just:
# >
# > _wallet fraction_ is _percentage_ of _wallet value_
# >
# > _target balance_ is _wallet fraction_ divided by _token price_
#
class PercentageTargetBalance(TargetBalance):
    def __init__(self, symbol: str, target_percentage: Decimal) -> None:
        super().__init__(symbol)
        self.target_fraction = target_percentage / 100

    def resolve(
        self, instrument: Instrument, current_price: Decimal, total_value: Decimal
    ) -> InstrumentValue:
        target_value = total_value * self.target_fraction
        target_size = target_value / current_price
        return InstrumentValue(instrument, target_size)

    def __str__(self) -> str:
        return f"""« PercentageTargetBalance [{self.target_fraction * 100}% {self.symbol}] »"""


# # 🥭 parse_target_balance function
#
# `argparse` handler for `TargetBalance` parsing. Can be used like:
# parser.add_argument("--target", type=entropy.parse_target_balance, action="append", required=True,
#                     help="token symbol plus target value or percentage, separated by a colon (e.g. 'ETH:2.5')")
#
def parse_target_balance(to_parse: str) -> TargetBalance:
    try:
        symbol, value = to_parse.split(":")
    except Exception as exception:
        raise Exception(f"Could not parse target balance '{to_parse}'") from exception

    # The value we have may be an int (like 27), a fraction (like 0.1) or a percentage
    # (like 25%). In all cases we want the number as a number, but we also want to know if
    # we have a percent or not
    values = value.split("%")
    numeric_value_string = values[0]
    try:
        numeric_value = Decimal(numeric_value_string)
    except Exception as exception:
        raise Exception(
            f"Could not parse '{numeric_value_string}' as a decimal number. It should be formatted as a decimal number, e.g. '2.345', with no surrounding spaces."
        ) from exception

    if len(values) > 2:
        raise Exception(
            f"Could not parse '{value}' as a decimal percentage. It should be formatted as a decimal number followed by a percentage sign, e.g. '30%', with no surrounding spaces."
        )

    if len(values) == 1:
        return FixedTargetBalance(symbol, numeric_value)
    else:
        return PercentageTargetBalance(symbol, numeric_value)


# # 🥭 parse_fixed_target_balance function
#
# `argparse` handler for `TargetBalance` parsing. Can only be used for `FixedTargetBalance`s - will raise an
# error if a `PercentageTargetBalance` is attempted. This is useful for circumstances where percentage
# balance targets aren't allowed.
#
# Can be used like:
# parser.add_argument("--target", type=entropy.parse_fixed_target_balance, action="append", required=True,
#                     help="token symbol plus target value or percentage, separated by a colon (e.g. 'ETH:2.5')")
#
def parse_fixed_target_balance(to_parse: str) -> TargetBalance:
    try:
        symbol, value = to_parse.split(":")
    except Exception as exception:
        raise Exception(f"Could not parse target balance '{to_parse}'") from exception

    # The value we have may be an int (like 27)or  a fraction (like 0.1). In all cases we want the number
    # as a number, but we also want to know if we have a percent or not so we can raise an exception.
    values = value.split("%")
    if len(values) > 1:
        raise Exception(
            f"Could not parse '{value}' as a decimal target. (Percentage targets are not allowed in this context.)"
        )

    numeric_value_string = values[0]
    try:
        numeric_value = Decimal(numeric_value_string)
    except Exception as exception:
        raise Exception(
            f"Could not parse '{numeric_value_string}' as a decimal number. It should be formatted as a decimal number, e.g. '2.345', with no surrounding spaces."
        ) from exception

    return FixedTargetBalance(symbol, numeric_value)


# # 🥭 sort_changes_for_trades function
#
# It's important to process SELLs first, so we have enough funds in the quote balance for the
# BUYs.
#
# It looks like this function takes size into account, but it doesn't really - 2 ETH is
# smaller than 1 BTC (for now?) but the value 2 will be treated as bigger than 1. We don't
# really care that much as long as we have SELLs before BUYs. (We could, later, take price
# into account for this sorting but we don't need to now so we don't.)
#
def sort_changes_for_trades(
    changes: typing.Sequence[InstrumentValue],
) -> typing.Sequence[InstrumentValue]:
    return sorted(changes, key=lambda change: change.value)


# # 🥭 calculate_required_balance_changes function
#
# Takes a list of current balances, and a list of desired balances, and returns the list of changes required to get us to the desired balances.
#
def calculate_required_balance_changes(
    current_balances: typing.Sequence[InstrumentValue],
    desired_balances: typing.Sequence[InstrumentValue],
) -> typing.Sequence[InstrumentValue]:
    changes: typing.List[InstrumentValue] = []
    for desired in desired_balances:
        current = InstrumentValue.find_by_token(current_balances, desired.token)
        change = InstrumentValue(desired.token, desired.value - current.value)
        changes += [change]

    return changes


# # 🥭 FilterSmallChanges class
#
# Allows us to filter out changes that aren't worth the effort.
#
# For instance, if our desired balance requires changing less than 1% of our total balance,
# it may not be worth bothering with right not.
#
# Calculations are based on the total wallet balance, rather than the magnitude of the
# change per-token, because a change of 0.01 of one token may be worth more than a change
# of 10 in another token. Normalising values to our wallet balance makes these changes
# easier to reason about.
#
class FilterSmallChanges:
    def __init__(
        self,
        action_threshold: Decimal,
        balances: typing.Sequence[InstrumentValue],
        prices: typing.Sequence[InstrumentValue],
    ) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.prices: typing.Dict[str, InstrumentValue] = {}
        total = Decimal(0)
        for balance in balances:
            price = InstrumentValue.find_by_token(prices, balance.token)
            self.prices[price.token.symbol] = price
            total += price.value * balance.value
        self.total_balance = total
        self.action_threshold_value = total * action_threshold
        self._logger.info(
            f"Wallet total balance of {total:,.8f} gives action threshold: {self.action_threshold_value:,.8f}"
        )

    def allow(self, token_value: InstrumentValue) -> bool:
        price = self.prices[token_value.token.symbol]
        value = price.value * token_value.value
        absolute_value = value.copy_abs()
        result = absolute_value > self.action_threshold_value

        self._logger.info(
            f"Worth doing? {result}. {token_value.token.name} trade is worth: {absolute_value:,.8f}, threshold is: {self.action_threshold_value:,.8f}."
        )
        return result


# # 🥭 WalletBalancers
#
# We want two types of this class:
# * 'null' implementation that adheres to the interface but doesn't do anything, and
# * 'live' implementations that actually do the balancing.
#
# This allows us to have code that implements logic including wallet balancing, without
# having to worry about whether the user wants to re-balance or not - we can just plug
# in the 'null' variant and the logic all still works.
#
# To have this work we define an abstract base class `WalletBalancer` which defines the
# interface, then a `NullWalletBalancer` which adheres to this interface but doesn't
# perform any action, and finally the real `LiveWalletBalancer` which can perform the
# balancing action.
#

# # 🥭 WalletBalancer class
#
# This is the abstract class which defines the interface.
#
class WalletBalancer(metaclass=abc.ABCMeta):
    def __init__(self) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)

    @abc.abstractmethod
    def balance(
        self, context: Context, prices: typing.Sequence[InstrumentValue]
    ) -> None:
        raise NotImplementedError(
            "WalletBalancer.balance() is not implemented on the base type."
        )


# # 🥭 NullWalletBalancer class
#
# This is the 'empty', 'no-op', 'dry run' wallet balancer which doesn't do anything but
# which can be plugged into algorithms that may want balancing logic.
#
class NullWalletBalancer(WalletBalancer):
    def __init__(self) -> None:
        super().__init__()

    def balance(
        self, context: Context, prices: typing.Sequence[InstrumentValue]
    ) -> None:
        pass


# # 🥭 LiveWalletBalancer class
#
# This is the high-level class that does much of the work.
#
class LiveWalletBalancer(WalletBalancer):
    def __init__(
        self,
        context: Context,
        wallet: Wallet,
        account: Account,
        quote_token: Token,
        targets: typing.Sequence[TargetBalance],
        action_threshold: Decimal,
        max_slippage: Decimal,
        dry_run: bool,
    ) -> None:
        super().__init__()
        self.context: Context = context
        self.wallet: Wallet = wallet
        self.account: Account = account
        self.quote_token: Token = quote_token
        self.targets: typing.Sequence[TargetBalance] = targets
        self.action_threshold: Decimal = action_threshold
        self.max_slippage: Decimal = max_slippage
        self.dry_run: bool = dry_run

    def balance(
        self, context: Context, prices: typing.Sequence[InstrumentValue]
    ) -> None:
        padding = "\n    "

        def balances_report(balances: typing.Sequence[InstrumentValue]) -> str:
            return padding.join(list([f"{bal}" for bal in balances]))

        tokens: typing.List[Token] = []
        for target_balance in self.targets:
            token = context.instrument_lookup.find_by_symbol(target_balance.symbol)
            if token is None:
                raise Exception(
                    f"Could not find details of token {target_balance.symbol}."
                )
            tokens += [Token.ensure(token)]
        tokens += [self.quote_token]

        balances = self._fetch_balances(context, tokens)
        total_value = Decimal(0)
        for bal in balances:
            price = InstrumentValue.find_by_token(prices, bal.token)
            value = bal.value * price.value
            total_value += value
        self._logger.info(f"Starting balances: {padding}{balances_report(balances)}")
        total_token_value: InstrumentValue = InstrumentValue(
            self.quote_token, total_value
        )
        self._logger.info(f"Total: {total_token_value}")

        resolved_targets: typing.List[InstrumentValue] = []
        for target in self.targets:
            price = InstrumentValue.find_by_symbol(prices, target.symbol)
            resolved_targets += [target.resolve(price.token, price.value, total_value)]

        balance_changes = calculate_required_balance_changes(balances, resolved_targets)
        self._logger.info(
            f"Desired balance changes: {padding}{balances_report(balance_changes)}"
        )

        dont_bother = FilterSmallChanges(self.action_threshold, balances, prices)
        filtered_changes = list(filter(dont_bother.allow, balance_changes))
        self._logger.info(
            f"Filtered balance changes: {padding}{balances_report(filtered_changes)}"
        )
        if len(filtered_changes) == 0:
            self._logger.info("No balance changes to make.")
            return

        sorted_changes = sort_changes_for_trades(filtered_changes)
        self._make_changes(sorted_changes)
        updated_balances = self._fetch_balances(context, tokens)
        self._logger.info(
            f"Finishing balances: {padding}{balances_report(updated_balances)}"
        )

    def _make_changes(self, balance_changes: typing.Sequence[InstrumentValue]) -> None:
        quote = self.quote_token.symbol
        for change in balance_changes:
            market_symbol = f"serum:{change.token.symbol}/{quote}"
            ops: MarketOperations = operations(
                self.context, self.wallet, self.account, market_symbol, self.dry_run
            )
            if change.value < 0:
                ops.market_sell(change.value.copy_abs(), self.max_slippage)
            else:
                ops.market_buy(change.value.copy_abs(), self.max_slippage)

    def _fetch_balances(
        self, context: Context, tokens: typing.Sequence[Token]
    ) -> typing.Sequence[InstrumentValue]:
        balances: typing.List[InstrumentValue] = []
        for token in tokens:
            balance = InstrumentValue.fetch_total_value(
                context, self.wallet.address, token
            )
            balances += [balance]

        return balances


# # 🥭 LiveAccountBalancer class
#
# This is the high-level class that does much of the work.
#
class LiveAccountBalancer(WalletBalancer):
    def __init__(
        self,
        context: Context,
        wallet: Wallet,
        account: Account,
        group: Group,
        targets: typing.Sequence[TargetBalance],
        action_threshold: Decimal,
        max_slippage: Decimal,
        dry_run: bool,
    ) -> None:
        super().__init__()
        self.context: Context = context
        self.wallet: Wallet = wallet
        self.account: Account = account
        self.group: Group = group
        self.targets: typing.Sequence[TargetBalance] = targets
        self.action_threshold: Decimal = action_threshold
        self.max_slippage: Decimal = max_slippage
        self.dry_run: bool = dry_run

    def balance(
        self, context: Context, prices: typing.Sequence[InstrumentValue]
    ) -> None:
        padding = "\n    "

        def balances_report(balances: typing.Sequence[InstrumentValue]) -> str:
            return padding.join(list([f"{bal}" for bal in balances]))

        balances = [basket_token.net_value for basket_token in self.account.base_slots]
        total_value = Decimal(0)
        for bal in balances:
            price = InstrumentValue.find_by_token(prices, bal.token)
            value = bal.value * price.value
            total_value += value
        self._logger.info(f"Starting balances: {padding}{balances_report(balances)}")
        quote_token: Token = self.account.shared_quote_token
        total_token_value: InstrumentValue = InstrumentValue(quote_token, total_value)
        self._logger.info(f"Total: {total_token_value}")
        resolved_targets: typing.List[InstrumentValue] = []
        for target in self.targets:
            price = InstrumentValue.find_by_symbol(prices, target.symbol)
            resolved_targets += [target.resolve(price.token, price.value, total_value)]

        balance_changes = calculate_required_balance_changes(balances, resolved_targets)
        self._logger.info(
            f"Desired balance changes: {padding}{balances_report(balance_changes)}"
        )

        dont_bother = FilterSmallChanges(self.action_threshold, balances, prices)
        filtered_changes = list(filter(dont_bother.allow, balance_changes))
        self._logger.info(
            f"Worthwhile balance changes: {padding}{balances_report(filtered_changes)}"
        )
        if len(filtered_changes) == 0:
            self._logger.info("No balance changes to make.")
            return

        sorted_changes = sort_changes_for_trades(filtered_changes)
        self._make_changes(sorted_changes)

        updated_account: Account = Account.load(
            context, self.account.address, self.group
        )
        updated_balances = [
            basket_token.net_value for basket_token in updated_account.base_slots
        ]
        self._logger.info(
            f"Finishing balances: {padding}{balances_report(updated_balances)}"
        )

    def _make_changes(self, balance_changes: typing.Sequence[InstrumentValue]) -> None:
        quote = self.account.shared_quote_token.symbol
        for change in balance_changes:
            market_symbol = f"spot:{change.token.symbol}/{quote}"
            ops: MarketOperations = operations(
                self.context, self.wallet, self.account, market_symbol, self.dry_run
            )
            if change.value < 0:
                ops.market_sell(change.value.copy_abs(), self.max_slippage)
            else:
                ops.market_buy(change.value.copy_abs(), self.max_slippage)



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import enum


# # 🥭 InstructionType enum
#
# This `enum` encapsulates all current Entropy Market instruction variants.
#


class InstructionType(enum.IntEnum):
    InitEntropyGroup = 0
    InitMarginAccount = 1
    Deposit = 2
    Withdraw = 3
    AddSpotMarket = 4
    AddToBasket = 5
    Borrow = 6
    CachePrices = 7
    CacheRootBanks = 8
    PlaceSpotOrder = 9
    AddOracle = 10
    AddPerpMarket = 11
    PlacePerpOrder = 12
    CancelPerpOrderByClientId = 13
    CancelPerpOrder = 14
    ConsumeEvents = 15
    CachePerpMarkets = 16
    UpdateFunding = 17
    SetOracle = 18
    SettleFunds = 19
    CancelSpotOrder = 20
    UpdateRootBank = 21
    SettlePnl = 22
    SettleBorrow = 23
    ForceCancelSpotOrders = 24
    ForceCancelPerpOrders = 25
    LiquidateTokenAndToken = 26
    LiquidateTokenAndPerp = 27
    LiquidatePerpMarket = 28
    SettleFees = 29
    ResolvePerpBankruptcy = 30
    ResolveTokenBankruptcy = 31
    InitSpotOpenOrders = 32
    RedeemMngo = 33
    AddEntropyAccountInfo = 34
    DepositMsrm = 35
    WithdrawMsrm = 36
    ChangePerpMarketParams = 37
    SetGroupAdmin = 38
    CancelAllPerpOrders = 39
    ForceSettleQuotePositions = 40
    PlaceSpotOrder2 = 41
    InitAdvancedOrders = 42
    AddPerpTriggerOrder = 43
    RemoveAdvancedOrder = 44
    ExecutePerpTriggerOrder = 45
    CreatePerpMarket = 46
    ChangePerpMarketParams2 = 47
    UpdateMarginBasket = 48
    ChangeMaxEntropyAccounts = 49
    CloseEntropyAccount = 50
    CloseSpotOpenOrders = 51
    CloseAdvancedOrders = 52
    CreateDustAccount = 53
    ResolveDust = 54
    CreateEntropyAccount = 55
    UpgradeEntropyAccountV0V1 = 56
    CancelPerpOrderSide = 57
    SetDelegate = 58
    ChangeSpotMarketParams = 59
    CreateSpotOpenOrders = 60
    ChangeReferralFeeParams = 61
    SetReferrerMemory = 62
    RegisterReferrerId = 63
    PlacePerpOrder2 = 64

    def __str__(self) -> str:
        return self.name



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import enum
import pandas
import pyserum.enums
import typing

from dataclasses import dataclass
from datetime import datetime, timedelta
from decimal import Decimal
from pyserum.market.types import Order as PySerumOrder
from solana.publickey import PublicKey

from .constants import SYSTEM_PROGRAM_ADDRESS
from .datetimes import utc_now, datetime_from_timestamp
from .lotsizeconverter import LotSizeConverter


# # 🥭 Orders
#
# This file holds some basic common orders data types.
#


# # 🥭 Side enum
#
# Is an order a Buy or a Sell?
#
class Side(enum.Enum):
    # We use strings here so that argparse can work with these as parameters.
    BUY = "BUY"
    SELL = "SELL"

    @staticmethod
    def from_value(value: pyserum.enums.Side) -> "Side":
        converted: pyserum.enums.Side = pyserum.enums.Side(int(value))
        return Side.BUY if converted == pyserum.enums.Side.BUY else Side.SELL

    def to_serum(self) -> pyserum.enums.Side:
        return pyserum.enums.Side.BUY if self == Side.BUY else pyserum.enums.Side.SELL

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 OrderType enum
#
# 3 order types are supported: Limit (most common), IOC (immediate or cancel - not placed on the order book
# so if it doesn't get filled immediately it is cancelled), and Post Only (only ever places orders on the
# orderbook - if this would be filled immediately without being placed on the order book it is cancelled).
#
class OrderType(enum.Enum):
    # We use strings here so that argparse can work with these as parameters.
    UNKNOWN = "UNKNOWN"
    LIMIT = "LIMIT"
    IOC = "IOC"
    POST_ONLY = "POST_ONLY"
    MARKET = "MARKET"
    POST_ONLY_SLIDE = "POST_ONLY_SLIDE"

    @staticmethod
    def from_value(value: Decimal) -> "OrderType":
        # From: https://github.com/blockworks-foundation/entropy-v3/blob/0c4d26e3e32821d871c5e5986edafbf694a44137/program/src/matching.rs#L212
        # pub enum OrderType {
        #     Limit = 0,
        #     ImmediateOrCancel = 1,
        #     PostOnly = 2,
        #     Market = 3,
        #     PostOnlySlide = 4, // ***
        # }
        if value == 0:
            return OrderType.LIMIT
        elif value == 1:
            return OrderType.IOC
        elif value == 2:
            return OrderType.POST_ONLY
        elif value == 3:
            return OrderType.MARKET
        elif value == 4:
            return OrderType.POST_ONLY_SLIDE

        return OrderType.LIMIT

    def to_serum(self) -> pyserum.enums.OrderType:
        if self == OrderType.IOC:
            return pyserum.enums.OrderType.IOC
        elif self == OrderType.POST_ONLY:
            return pyserum.enums.OrderType.POST_ONLY
        elif self == OrderType.POST_ONLY_SLIDE:
            # Best we can do in this situation
            return pyserum.enums.OrderType.POST_ONLY
        else:
            return pyserum.enums.OrderType.LIMIT

    def to_perp(self) -> int:
        # From: https://github.com/blockworks-foundation/entropy-v3/blob/0c4d26e3e32821d871c5e5986edafbf694a44137/program/src/matching.rs#L212
        # pub enum OrderType {
        #     Limit = 0,
        #     ImmediateOrCancel = 1,
        #     PostOnly = 2,
        #     Market = 3,
        #     PostOnlySlide = 4, // ***
        # }
        if self == OrderType.LIMIT:
            return 0
        elif self == OrderType.IOC:
            return 1
        elif self == OrderType.POST_ONLY:
            return 2
        elif self == OrderType.MARKET:
            return 3
        elif self == OrderType.POST_ONLY_SLIDE:
            return 4
        else:
            return -1

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 Order named tuple
#
# A package that encapsulates common information about an order.
#
@dataclass(frozen=True)
class Order:
    DefaultMatchLimit: typing.ClassVar[int] = 20
    NoExpiration: typing.ClassVar[datetime] = datetime_from_timestamp(0)
    id: int
    client_id: int
    owner: PublicKey
    side: Side
    price: Decimal
    quantity: Decimal
    order_type: OrderType
    reduce_only: bool = False
    timestamp: typing.Optional[datetime] = None
    expiration: datetime = NoExpiration
    match_limit: int = DefaultMatchLimit

    @staticmethod
    def read_sequence_number(id: int) -> Decimal:
        id_bytes = id.to_bytes(16, byteorder="little", signed=False)
        low_order = id_bytes[:8]
        return Decimal(int.from_bytes(low_order, "little", signed=False))

    @staticmethod
    def read_price(id: int) -> Decimal:
        id_bytes = id.to_bytes(16, byteorder="little", signed=False)
        high_order = id_bytes[8:]
        return Decimal(int.from_bytes(high_order, "little", signed=False))

    def _emoji_marker_at(self, cutoff: datetime) -> str:
        if self.is_expired_at(cutoff):
            return "⛔"
        elif self.expiration != Order.NoExpiration:
            return "⏰"
        return "📌"

    def is_expired_at(self, cutoff: typing.Optional[datetime]) -> bool:
        if (
            (cutoff is None)
            or (self.expiration == Order.NoExpiration)
            or (self.expiration > cutoff)
        ):
            return False
        return True

    # Returns an identical order with the provided values changed.
    def with_update(
        self,
        id: typing.Optional[int] = None,
        client_id: typing.Optional[int] = None,
        owner: typing.Optional[PublicKey] = None,
        side: typing.Optional[Side] = None,
        price: typing.Optional[Decimal] = None,
        quantity: typing.Optional[Decimal] = None,
        order_type: typing.Optional[OrderType] = None,
        reduce_only: typing.Optional[bool] = None,
        timestamp: typing.Optional[datetime] = None,
        expiration: typing.Optional[datetime] = None,
        match_limit: typing.Optional[int] = None,
    ) -> "Order":
        return Order(
            id=id if id is not None else self.id,
            client_id=client_id if client_id is not None else self.client_id,
            owner=owner if owner is not None else self.owner,
            side=side if side is not None else self.side,
            price=price if price is not None else self.price,
            quantity=quantity if quantity is not None else self.quantity,
            order_type=order_type if order_type is not None else self.order_type,
            reduce_only=reduce_only if reduce_only is not None else self.reduce_only,
            timestamp=timestamp if timestamp is not None else self.timestamp,
            expiration=expiration if expiration is not None else self.expiration,
            match_limit=match_limit if match_limit is not None else self.match_limit,
        )

    @staticmethod
    def from_serum_order(serum_order: PySerumOrder) -> "Order":
        price = Decimal(serum_order.info.price)
        quantity = Decimal(serum_order.info.size)
        side = Side.from_value(serum_order.side)
        order = Order(
            id=serum_order.order_id,
            side=side,
            price=price,
            quantity=quantity,
            client_id=serum_order.client_id,
            owner=serum_order.open_order_address,
            order_type=OrderType.UNKNOWN,
        )
        return order

    @staticmethod
    def from_values(
        side: Side,
        price: Decimal,
        quantity: Decimal,
        order_type: OrderType = OrderType.UNKNOWN,
        id: int = 0,
        client_id: int = 0,
        owner: PublicKey = SYSTEM_PROGRAM_ADDRESS,
        reduce_only: bool = False,
        timestamp: typing.Optional[datetime] = None,
        expiration: datetime = NoExpiration,
        match_limit: int = 20,
    ) -> "Order":
        order = Order(
            id=id,
            side=side,
            price=price,
            quantity=quantity,
            client_id=client_id,
            owner=owner,
            order_type=order_type,
            reduce_only=reduce_only,
            timestamp=timestamp,
            expiration=expiration,
            match_limit=match_limit,
        )
        return order

    @staticmethod
    def from_ids(
        id: int,
        client_id: int,
        side: Side = Side.BUY,
        price: Decimal = Decimal(0),
        quantity: Decimal = Decimal(0),
    ) -> "Order":
        return Order(
            id=id,
            client_id=client_id,
            owner=SYSTEM_PROGRAM_ADDRESS,
            side=side,
            price=price,
            quantity=quantity,
            order_type=OrderType.UNKNOWN,
            reduce_only=False,
            timestamp=None,
            expiration=Order.NoExpiration,
            match_limit=20,
        )

    @staticmethod
    def build_absolute_expiration(
        expire_seconds: typing.Optional[Decimal],
    ) -> datetime:
        if expire_seconds is None or expire_seconds <= Decimal(0):
            return Order.NoExpiration

        return utc_now() + timedelta(seconds=float(expire_seconds))

    def __str__(self) -> str:
        owner: str = ""
        if self.owner != SYSTEM_PROGRAM_ADDRESS:
            owner = f"[{self.owner}] "
        order_type: str = ""
        if self.order_type != OrderType.UNKNOWN:
            order_type = f" {self.order_type}"
        marker = self._emoji_marker_at(utc_now())
        return f"« Order {marker} {owner}{self.side} {self.quantity:,.8f} at {self.price:.8f} [ID: {self.id} / {self.client_id}]{order_type}{' reduceOnly' if self.reduce_only else ''} »"

    def __repr__(self) -> str:
        return f"{self}"


class OrderBook:
    def __init__(
        self,
        symbol: str,
        lot_size_converter: LotSizeConverter,
        bids: typing.Sequence[Order],
        asks: typing.Sequence[Order],
    ) -> None:
        self.symbol: str = symbol
        self.__lot_size_converter: LotSizeConverter = lot_size_converter
        self.__bids: typing.Sequence[Order] = []
        self.__asks: typing.Sequence[Order] = []
        self.bids = bids
        self.asks = asks

    @property
    def bids(self) -> typing.Sequence[Order]:
        return self.bids_at(cutoff=utc_now())

    @bids.setter
    def bids(self, bids: typing.Sequence[Order]) -> None:
        """Sort bids high to low, so best bid is at index 0"""
        bids_list: typing.List[Order] = list(bids)
        bids_list.sort(key=lambda order: order.id, reverse=True)
        self.__bids = bids_list

    @property
    def asks(self) -> typing.Sequence[Order]:
        return self.asks_at(cutoff=utc_now())

    @asks.setter
    def asks(self, asks: typing.Sequence[Order]) -> None:
        """Sets asks low to high, so best ask is at index 0"""
        asks_list: typing.List[Order] = list(asks)
        asks_list.sort(key=lambda order: order.id)
        self.__asks = asks_list

    # The top bid is the highest price someone is willing to pay to BUY
    @property
    def top_bid(self) -> typing.Optional[Order]:
        return self.top_bid_at(cutoff=utc_now())

    # The top ask is the lowest price someone is willing to pay to SELL
    @property
    def top_ask(self) -> typing.Optional[Order]:
        return self.top_ask_at(cutoff=utc_now())

    # The mid price is halfway between the best bid and best ask.
    @property
    def mid_price(self) -> typing.Optional[Decimal]:
        return self.mid_price_at(cutoff=utc_now())

    @property
    def spread(self) -> Decimal:
        return self.spread_at(cutoff=utc_now())

    def bids_at(
        self, cutoff: typing.Optional[datetime] = None
    ) -> typing.Sequence[Order]:
        return list([o for o in self.__bids if not o.is_expired_at(cutoff)])

    def asks_at(
        self, cutoff: typing.Optional[datetime] = None
    ) -> typing.Sequence[Order]:
        return list([o for o in self.__asks if not o.is_expired_at(cutoff)])

    def orders_at(
        self, cutoff: typing.Optional[datetime] = None
    ) -> typing.Sequence[Order]:
        return [*self.bids_at(cutoff), *self.asks_at(cutoff)]

    # The top bid is the highest price someone is willing to pay to BUY
    def top_bid_at(
        self, cutoff: typing.Optional[datetime] = None
    ) -> typing.Optional[Order]:
        bids = self.bids_at(cutoff)
        if bids and len(bids) > 0:
            # Top-of-book is always at index 0 for us.
            return bids[0]
        return None

    # The top ask is the lowest price someone is willing to pay to SELL
    def top_ask_at(
        self, cutoff: typing.Optional[datetime] = None
    ) -> typing.Optional[Order]:
        asks = self.asks_at(cutoff)
        if asks and len(asks) > 0:
            # Top-of-book is always at index 0 for us.
            return asks[0]
        return None

    # The mid price is halfway between the best bid and best ask.
    def mid_price_at(
        self, cutoff: typing.Optional[datetime] = None
    ) -> typing.Optional[Decimal]:
        top_bid = self.top_bid_at(cutoff)
        top_ask = self.top_ask_at(cutoff)
        if top_bid is not None and top_ask is not None:
            return (top_bid.price + top_ask.price) / 2
        elif top_bid is not None:
            return top_bid.price
        elif top_ask is not None:
            return top_ask.price
        return None

    def spread_at(self, cutoff: typing.Optional[datetime] = None) -> Decimal:
        top_ask = self.top_ask_at(cutoff)
        top_bid = self.top_bid_at(cutoff)
        if top_ask is None or top_bid is None:
            return Decimal(0)
        else:
            return top_ask.price - top_bid.price

    def all_orders_for_owner(
        self, owner_address: PublicKey, cutoff: typing.Optional[datetime] = None
    ) -> typing.Sequence[Order]:
        return list([o for o in self.orders_at(cutoff) if o.owner == owner_address])

    def to_dataframe(self) -> pandas.DataFrame:
        column_mapper = {
            "id": "Id",
            "client_id": "ClientId",
            "owner": "Owner",
            "side": "Side",
            "price": "Price",
            "quantity": "Quantity",
            "timestamp": "Timestamp",
            "expiration": "Expiration",
        }

        frame: pandas.DataFrame = pandas.DataFrame([*reversed(self.bids), *self.asks])
        frame = frame.drop(["order_type"], axis=1)
        frame = frame.rename(mapper=column_mapper, axis=1, copy=True)
        frame["Price"] = pandas.to_numeric(frame["Price"])
        frame["QuantityLots"] = frame["Quantity"].apply(
            self.__lot_size_converter.base_size_number_to_lots
        )
        frame["Quantity"] = pandas.to_numeric(frame["Quantity"])
        frame["PriceLots"] = pandas.to_numeric(frame["Id"].apply(Order.read_price))
        frame["SequenceNumber"] = frame["Id"].apply(Order.read_sequence_number)

        return frame

    def to_l1_dataframe(self) -> pandas.DataFrame:
        frame: pandas.DataFrame = self.to_l2_dataframe()
        buys = frame[(frame["Side"] == Side.BUY)]
        sells = frame[(frame["Side"] == Side.SELL)]
        top = frame.loc[[buys["PriceLots"].idxmax(), sells["PriceLots"].idxmin()]]
        return typing.cast(pandas.DataFrame, top)

    def to_l2_dataframe(self) -> pandas.DataFrame:
        frame: pandas.DataFrame = self.to_dataframe()

        return frame.groupby("Price").agg(
            {
                "PriceLots": "first",
                "Side": "first",
                "Quantity": "sum",
                "QuantityLots": "sum",
            }
        )

    def to_l3_dataframe(self) -> pandas.DataFrame:
        return self.to_dataframe()

    def __str__(self) -> str:
        def _order_to_str(order: Order) -> str:
            marker = order._emoji_marker_at(utc_now())
            quantity = f"{order.quantity:,.8f}"
            price = f"{order.price:,.8f}"
            return f"{marker} {order.side} {quantity:>20} at {price:>20}"

        orders_to_show = 5
        lines = []
        for counter in range(orders_to_show):
            bid = _order_to_str(self.bids[counter]) if len(self.bids) > counter else ""
            ask = _order_to_str(self.asks[counter]) if len(self.asks) > counter else ""
            lines += [f"{bid:50} :: {ask}"]

        text = "\n\t".join(lines)
        spread_description = "N/A"
        if self.spread != 0 and self.top_bid is not None:
            spread_percentage = self.spread / self.top_bid.price
            spread_description = f"{self.spread:,.8f}, {spread_percentage:,.3%}"
        return f"« OrderBook {self.symbol} [spread: {spread_description}]\n\t{text}\n»"

    def __repr__(self) -> str:
        return f"{self}"



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import base58
import logging
import traceback
import typing

from datetime import datetime
from decimal import Decimal
from solana.publickey import PublicKey

from .context import Context
from .datetimes import datetime_from_timestamp
from .instructiontype import InstructionType
from .instrumentvalue import InstrumentValue
from .logmessages import expand_log_messages
from .entropyinstruction import EntropyInstruction
from .layouts import layouts
from .ownedinstrumentvalue import OwnedInstrumentValue
from .text import indent_collection_as_str, indent_item_by


# # 🥭 TransactionScout
#
# ## Transaction Indices
#
# Transactions come with a large account list.
#
# Instructions, individually, take accounts.
#
# The accounts instructions take are listed in the the transaction's list of accounts.
#
# The instruction data therefore doesn't need to specify account public keys, only the
# index of those public keys in the main transaction's list.
#
# So, for example, if an instruction uses 3 accounts, the instruction data could say
# [3, 2, 14], meaning the first account it uses is index 3 in the whole transaction account
# list, the second is index 2 in the whole transaction account list, the third is index 14
# in the whole transaction account list.
#
# This complicates figuring out which account is which for a given instruction, especially
# since some of the accounts (like the sender/signer account) can appear at different
# indices depending on which instruction is being used.
#
# We keep a few static dictionaries here to allow us to dereference important accounts per
# type.
#
# In addition, we dereference the accounts for each instruction when we instantiate each
# `EntropyInstruction`, so users of `EntropyInstruction` don't need to worry about
# these details.
#


# # 🥭 TransactionScout class
#
class TransactionScout:
    def __init__(
        self,
        timestamp: datetime,
        signatures: typing.Sequence[str],
        succeeded: bool,
        group_name: str,
        accounts: typing.Sequence[PublicKey],
        instructions: typing.Sequence[EntropyInstruction],
        messages: typing.Sequence[str],
        pre_token_balances: typing.Sequence[OwnedInstrumentValue],
        post_token_balances: typing.Sequence[OwnedInstrumentValue],
    ) -> None:
        self.timestamp: datetime = timestamp
        self.signatures: typing.Sequence[str] = signatures
        self.succeeded: bool = succeeded
        self.group_name: str = group_name
        self.accounts: typing.Sequence[PublicKey] = accounts
        self.instructions: typing.Sequence[EntropyInstruction] = instructions
        self.messages: typing.Sequence[str] = messages
        self.pre_token_balances: typing.Sequence[
            OwnedInstrumentValue
        ] = pre_token_balances
        self.post_token_balances: typing.Sequence[
            OwnedInstrumentValue
        ] = post_token_balances

    @property
    def summary(self) -> str:
        result = "[Success]" if self.succeeded else "[Failed]"
        instructions = ", ".join(
            [ins.instruction_type.name for ins in self.instructions]
        )
        changes = OwnedInstrumentValue.changes(
            self.pre_token_balances, self.post_token_balances
        )

        in_tokens = []
        for ins in self.instructions:
            if ins.token_in_account is not None:
                in_tokens += [
                    OwnedInstrumentValue.find_by_owner(changes, ins.token_in_account)
                ]

        out_tokens = []
        for ins in self.instructions:
            if ins.token_out_account is not None:
                out_tokens += [
                    OwnedInstrumentValue.find_by_owner(changes, ins.token_out_account)
                ]

        changed_tokens = in_tokens + out_tokens
        changed_tokens_text = (
            ", ".join(
                [
                    f"{tok.token_value.value:,.8f} {tok.token_value.token.name}"
                    for tok in changed_tokens
                ]
            )
            or "None"
        )

        return f"« TransactionScout {result} {self.group_name} [{self.timestamp}] {instructions}: Token Changes: {changed_tokens_text}\n    {self.signatures} »"

    @property
    def sender(self) -> typing.Optional[PublicKey]:
        return self.instructions[0].sender if len(self.instructions) > 0 else None

    @property
    def group(self) -> typing.Optional[PublicKey]:
        return self.instructions[0].group if len(self.instructions) > 0 else None

    def has_any_instruction_of_type(self, instruction_type: InstructionType) -> bool:
        return any(
            map(lambda ins: ins.instruction_type == instruction_type, self.instructions)
        )

    @staticmethod
    def load_if_available(
        context: Context, signature: str
    ) -> typing.Optional["TransactionScout"]:
        transaction_details = context.client.get_confirmed_transaction(signature)
        if transaction_details is None:
            return None
        return TransactionScout.from_transaction_response(context, transaction_details)

    @staticmethod
    def load(context: Context, signature: str) -> "TransactionScout":
        tx = TransactionScout.load_if_available(context, signature)
        if tx is None:
            raise Exception(f"Transaction '{signature}' not found.")
        return tx

    @staticmethod
    def from_transaction_response(
        context: Context, response: typing.Dict[str, typing.Any]
    ) -> "TransactionScout":
        def balance_to_token_value(
            accounts: typing.Sequence[PublicKey], balance: typing.Dict[str, typing.Any]
        ) -> OwnedInstrumentValue:
            mint = PublicKey(balance["mint"])
            account = accounts[balance["accountIndex"]]
            amount = Decimal(balance["uiTokenAmount"]["amount"])
            decimals = Decimal(balance["uiTokenAmount"]["decimals"])
            divisor = Decimal(10) ** decimals
            value = amount / divisor
            token = context.instrument_lookup.find_by_mint_or_raise(mint)
            return OwnedInstrumentValue(account, InstrumentValue(token, value))

        try:
            succeeded = True if response["meta"]["err"] is None else False
            accounts = list(
                map(PublicKey, response["transaction"]["message"]["accountKeys"])
            )
            instructions: typing.List[EntropyInstruction] = []
            for instruction_data in response["transaction"]["message"]["instructions"]:
                instruction = entropy_instruction_from_response(
                    context, accounts, instruction_data
                )
                if instruction is not None:
                    instructions += [instruction]

            group_name = (
                context.lookup_group_name(instructions[0].group)
                if len(instructions) > 0
                else "No Group"
            )
            timestamp = datetime_from_timestamp(response["blockTime"])
            signatures = response["transaction"]["signatures"]
            raw_messages = response["meta"]["logMessages"]
            messages = expand_log_messages(raw_messages)
            pre_token_balances = list(
                map(
                    lambda bal: balance_to_token_value(accounts, bal),
                    response["meta"]["preTokenBalances"],
                )
            )
            post_token_balances = list(
                map(
                    lambda bal: balance_to_token_value(accounts, bal),
                    response["meta"]["postTokenBalances"],
                )
            )
            return TransactionScout(
                timestamp,
                signatures,
                succeeded,
                group_name,
                accounts,
                instructions,
                messages,
                pre_token_balances,
                post_token_balances,
            )
        except Exception as exception:
            signature = "Unknown"
            if (
                response
                and ("transaction" in response)
                and ("signatures" in response["transaction"])
                and len(response["transaction"]["signatures"]) > 0
            ):
                signature = ", ".join(response["transaction"]["signatures"])
            raise Exception(
                f"Exception fetching transaction '{signature}' - {traceback.format_exc()}",
                exception,
            )

    def __str__(self) -> str:
        def format_tokens(
            account_token_values: typing.Sequence[OwnedInstrumentValue],
        ) -> str:
            if len(account_token_values) == 0:
                return "None"
            return "\n        ".join([f"{atv}" for atv in account_token_values])

        instruction_names = ", ".join(
            [ins.instruction_type.name for ins in self.instructions]
        )
        signatures = indent_item_by(indent_collection_as_str(self.signatures), 2)
        accounts = indent_item_by(indent_collection_as_str(self.accounts), 2)
        messages = indent_item_by(indent_collection_as_str(self.messages), 2)
        instructions = indent_item_by(indent_collection_as_str(self.instructions), 2)
        changes = OwnedInstrumentValue.changes(
            self.pre_token_balances, self.post_token_balances
        )
        tokens_in = format_tokens(self.pre_token_balances)
        tokens_out = format_tokens(self.post_token_balances)
        token_changes = format_tokens(changes)
        return f"""« TransactionScout {self.timestamp}: {instruction_names}
    Sender:
        {self.sender}
    Succeeded:
        {self.succeeded}
    Group:
        {self.group_name} [{self.group}]
    Signatures:
        {signatures}
    Instructions:
        {instructions}
    Accounts:
        {accounts}
    Messages:
        {messages}
    Tokens In:
        {tokens_in}
    Tokens Out:
        {tokens_out}
    Token Changes:
        {token_changes}
»"""

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 fetch_all_recent_transaction_signatures function
#
def fetch_all_recent_transaction_signatures(context: Context) -> typing.Sequence[str]:
    all_fetched = False
    before = None
    signature_results: typing.List[str] = []
    while not all_fetched:
        signatures = context.client.get_confirmed_signatures_for_address2(
            context.group_address, before=before
        )
        signature_results += signatures
        if len(signatures) == 0:
            all_fetched = True
        else:
            before = signature_results[-1]

    return signature_results


def entropy_instruction_from_response(
    context: Context,
    all_accounts: typing.Sequence[PublicKey],
    instruction_data: typing.Dict[str, typing.Any],
) -> typing.Optional["EntropyInstruction"]:
    program_account_index = instruction_data["programIdIndex"]
    program_account: PublicKey = all_accounts[program_account_index]
    if program_account != context.entropy_program_address:
        # It's an instruction, it's just not a Entropy one.
        return None

    data = instruction_data["data"]
    instructions_account_indices = instruction_data["accounts"]

    decoded = base58.b58decode(data)
    initial = layouts.MANGO_INSTRUCTION_VARIANT_FINDER.parse(decoded)
    parser = layouts.InstructionParsersByVariant[initial.variant]
    if parser is None:
        logging.warning(
            f"Could not find instruction parser for variant {initial.variant} / {InstructionType(initial.variant)}."
        )
        return None

    # A whole bunch of accounts are listed for a transaction. Some (or all) of them apply
    # to this instruction. The instruction data gives the index of each account it uses,
    # in the order in which it uses them. So, for example, if it uses 3 accounts, the
    # instruction data could say [3, 2, 14], meaning the first account it uses is index 3
    # in the whole transaction account list, the second is index 2 in the whole transaction
    # account list, the third is index 14 in the whole transaction account list.
    accounts: typing.List[PublicKey] = []
    for index in instructions_account_indices:
        accounts += [all_accounts[index]]

    parsed = parser.parse(decoded)
    instruction_type = InstructionType(int(parsed.variant))

    return EntropyInstruction(
        program_account, instruction_type, decoded, parsed, accounts
    )



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://markets/) support is available at:
#   [Docs](https://docs.markets/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import requests
import re
import rx
import typing
import websocket

from decimal import Decimal
from rx.subject.subject import Subject

from ...context import Context
from ...datetimes import utc_now, datetime_from_timestamp
from ...loadedmarket import LoadedMarket
from ...observables import Disposable, DisposeWrapper
from ...oracle import (
    Oracle,
    OracleProvider,
    OracleSource,
    Price,
    SupportedOracleFeature,
)
from ...reconnectingwebsocket import ReconnectingWebsocket


# # 🥭 FTX
#
# This file contains code specific to the [Ftx Network](https://ftx.com/).
#
def _ftx_get_from_url(url: str) -> typing.Any:
    response = requests.get(url)
    response_values = response.json()
    if ("success" not in response_values) or (not response_values["success"]):
        raise Exception(f"Failed to get from FTX URL: {url} - {response_values}")
    return response_values["result"]


# # 🥭 FtxOracleConfidence constant
#
# FTX doesn't provide a confidence value.
#
FtxOracleConfidence: Decimal = Decimal(0)


# # 🥭 FtxOracle class
#
# Implements the `Oracle` abstract base class specialised to the Ftx Network.
#
class FtxOracle(Oracle):
    def __init__(self, market: LoadedMarket, ftx_symbol: str) -> None:
        name = f"Ftx Oracle for {market.fully_qualified_symbol} / {ftx_symbol}"
        super().__init__(name, market)
        self.market: LoadedMarket = market
        self.ftx_symbol: str = ftx_symbol
        features: SupportedOracleFeature = (
            SupportedOracleFeature.MID_PRICE | SupportedOracleFeature.TOP_BID_AND_OFFER
        )
        self.source: OracleSource = OracleSource("FTX", name, features, market)

    def fetch_price(self, context: Context) -> Price:
        result = _ftx_get_from_url(f"https://ftx.com/api/markets/{self.ftx_symbol}")
        bid = Decimal(result["bid"])
        ask = Decimal(result["ask"])
        price = Decimal(result["price"])

        return Price(
            self.source,
            utc_now(),
            self.market,
            bid,
            price,
            ask,
            FtxOracleConfidence,
        )

    def to_streaming_observable(self, _: Context) -> rx.core.typing.Observable[Price]:
        subject = Subject()

        def _on_item(data: typing.Dict[str, typing.Any]) -> None:
            if data["type"] == "update":
                bid = Decimal(data["data"]["bid"])
                ask = Decimal(data["data"]["ask"])
                mid = (bid + ask) / Decimal(2)
                time = data["data"]["time"]
                timestamp = datetime_from_timestamp(time)
                price = Price(
                    self.source,
                    timestamp,
                    self.market,
                    bid,
                    mid,
                    ask,
                    FtxOracleConfidence,
                )
                subject.on_next(price)

        def on_open(sock: websocket.WebSocketApp) -> None:
            sock.send(
                f"""{{"op": "subscribe", "channel": "ticker", "market": "{self.ftx_symbol}"}}"""
            )

        ws: ReconnectingWebsocket = ReconnectingWebsocket("wss://ftx.com/ws/", on_open)
        ws.item.subscribe(on_next=_on_item)  # type: ignore[call-arg]

        def subscribe(
            observer: rx.core.typing.Observer[Price],
            scheduler_: typing.Optional[rx.core.typing.Scheduler] = None,
        ) -> rx.core.typing.Disposable:
            subject.subscribe(observer, scheduler=scheduler_)  # type: ignore

            disposable = Disposable()
            disposable.add_disposable(DisposeWrapper(lambda: ws.close()))
            disposable.add_disposable(DisposeWrapper(lambda: subject.dispose()))

            return disposable

        price_observable = rx.core.observable.observable.Observable(subscribe)

        ws.open()

        return price_observable


# # 🥭 FtxOracleProvider class
#
# Implements the `OracleProvider` abstract base class specialised to the Ftx Network.
#


class FtxOracleProvider(OracleProvider):
    def __init__(self) -> None:
        super().__init__("Ftx Oracle Factory")

    def oracle_for_market(
        self, context: Context, market: LoadedMarket
    ) -> typing.Optional[Oracle]:
        symbol = self._market_symbol_to_ftx_symbol(market.symbol)
        return FtxOracle(market, symbol)

    def all_available_symbols(self, context: Context) -> typing.Sequence[str]:
        result = _ftx_get_from_url("https://ftx.com/api/markets")
        symbols: typing.List[str] = []
        for market in result:
            symbol: str = market["name"]
            if symbol.endswith("USD"):
                symbol = f"{symbol}C"
            symbols += [symbol]

        return symbols

    def _market_symbol_to_ftx_symbol(self, symbol: str) -> str:
        normalised = symbol.upper()
        return re.sub("USDC$", "USD", normalised)



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import abc
import random
import time


class IdGenerator(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def generate_id(self) -> int:
        raise NotImplementedError(
            "IdGenerator.generate_id() is not implemented on the base type."
        )


class RandomIdGenerator(IdGenerator):
    def generate_id(self) -> int:
        # 9223372036854775807 is sys.maxsize for 64-bit systems, with a bit_length of 63.\n",
        # We explicitly want to use a max of 64-bits though, so we use the number instead of\n",
        # sys.maxsize, which could be lower on 32-bit systems or higher on 128-bit systems.\n",
        return random.randrange(9223372036854775807)  # nosemgrep


class MonotonicIdGenerator(IdGenerator):
    def __init__(self) -> None:
        self.__last_generated_id: int = 0

    def generate_id(self) -> int:
        # After this discussion with Max on Discord (https://discord.com/channels/791995070613159966/818978757648842782/884751007656054804):
        #   can you generate monotonic ids?
        #   in case not the result wouldn't be different from what we have rn, which is random display
        #   so there's still a net benefit for changing the UI
        #   and if you could use the same id generation scheme (unix time in ms) it would even work well with the UI :slight_smile:
        #
        # We go with the time in milliseconds. We get the time in nanoseconds and divide it by 1,000,000 to get
        # the time in milliseconds.
        #
        # But there's more! Because this can be called in a burst, for, say, a dozen orders all within the same
        # millisecond. And using duplicate client order IDs would be Bad. So we keep track of the last one we
        # sent, and we just add one if we get an identical value.
        new_id: int = round(time.time_ns() / 1000000)
        if new_id <= self.__last_generated_id:
            new_id = self.__last_generated_id + 1
        self.__last_generated_id = new_id
        return new_id



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import base64
import base58
import typing
import zstandard

from solana.publickey import PublicKey


# # 🥭 Decoder
#
# This file contains some useful functions for decoding base64 and base58 data.

_decompressor: zstandard.ZstdDecompressor = zstandard.ZstdDecompressor()


# ## decode_binary() function
#
# A Solana binary data structure may come back as an array with the base64 or base58 encoded data, and a text moniker saying which encoding was used.
#
# For example:
# ```
# ['AwAAAAAAAACCaOmpoURMK6XHelGTaFawcuQ/78/15LAemWI8jrt3SRKLy2R9i60eclDjuDS8+p/ZhvTUd9G7uQVOYCsR6+BhmqGCiO6EPYP2PQkf/VRTvw7JjXvIjPFJy06QR1Cq1WfTonHl0OjCkyEf60SD07+MFJu5pVWNFGGEO/8AiAYfduaKdnFTaZEHPcK5Eq72WWHeHg2yIbBF09kyeOhlCJwOoG8O5SgpPV8QOA64ZNV4aKroFfADg6kEy/wWCdp3fv2B8WJgAAAAANVfH3HGtjwAAQAAAAAAAADr8cwFi9UOAAEAAAAAAAAAgfFiYAAAAABo3Dbz0L0oAAEAAAAAAAAAr8K+TvCjCwABAAAAAAAAAIHxYmAAAAAA49t5tVNZhwABAAAAAAAAAAmPtcB1zC8AAQAAAAAAAABIBGiCcyaEZdNhrTyeqUY692vOzzPdHaxAxguht3JQGlkzjtd05dX9LENHkl2z1XvUbTNKZlweypNRetmH0lmQ9VYQAHqylxZVK65gEg85g27YuSyvOBZAjJyRmYU9KdCO1D+4ehdPu9dQB1yI1uh75wShdAaFn2o4qrMYwq3SQQEAAAAAAAAAAiH1PPJKAuh6oGiE35aGhUQhFi/bxgKOudpFv8HEHNCFDy1uAqR6+CTQmradxC1wyyjL+iSft+5XudJWwSdi72NJGmyK96x7Obj/AgAAAAB8RjOEdJow6r9LMhIAAAAAGkNK4CXHh5M2st7PnwAAAE33lx1h8hPFD04AAAAAAAA8YRV3Oa309B2wGwAAAAAAOIlOLmkr6+r605n+AQAAAACgmZmZmZkZAQAAAAAAAAAAMDMzMzMzMwEAAAAAAAAA25D1XcAtRzSuuyx3U+X7aE9vM1EJySU9KprgL0LMJ/vat9+SEEUZuga7O5tTUrcMDYWDg+LYaAWhSQiN2fYk7aCGAQAAAAAAgIQeAAAAAAAA8gUqAQAAAAYGBgICAAAA', 'base64']
# ```
# Alternatively, it may just be a base58-encoded string.
#
# `decode_binary()` decodes the data properly based on which encoding was used.
def decode_binary(encoded: typing.Union[str, typing.Sequence[str]]) -> bytes:
    if isinstance(encoded, str):
        return base58.b58decode(encoded)
    elif encoded[1] == "base64":
        return base64.b64decode(encoded[0])
    elif encoded[1] == "base64+zstd":
        compressed = base64.b64decode(encoded[0])
        with _decompressor.stream_reader(compressed) as reader:
            return reader.read()
    else:
        return base58.b58decode(encoded[0])


# ## encode_binary() function
#
# Inverse of `decode_binary()`, this takes a binary list and encodes it (using base 64), then returns the encoded string and the string "base64" in an array.
#
def encode_binary(decoded: bytes) -> typing.Sequence[str]:
    return [base64.b64encode(decoded).decode(), "base64"]


# ## encode_key() function
#
# Encodes a `PublicKey` in the proper way for RPC calls.
def encode_key(key: PublicKey) -> str:
    return str(key)


# ## encode_int() function
#
# Encodes an `int` in the proper way for RPC calls.
def encode_int(value: int) -> str:
    return base58.b58encode_int(value).decode("ascii")
